--Match Function NB this is a separate package and may be removed to
--a separate file if required.  The test bench code is below the break.
--If this is removed then the util package must be compiled for each project

LIBRARY ieee;
use ieee.std_logic_1164.ALL;

package util is

 constant NO_WARNING: BOOLEAN := FALSE; -- default to emit warnings

-- Declare functions and procedure

  function MATCH (L, R: STD_ULOGIC) return BOOLEAN;
  function MATCH (L, R: STD_LOGIC_VECTOR) return BOOLEAN;
  function MATCH (L, R: STD_ULOGIC_VECTOR) return BOOLEAN;

end util;

package body util is


  -- support constants for MATCH:

  type BOOLEAN_TABLE is array(STD_ULOGIC, STD_ULOGIC) of BOOLEAN;

  constant MATCH_TABLE: BOOLEAN_TABLE := (
      --------------------------------------------------------------------------
      -- U      X      0      1      Z      W      L      H      -
      --------------------------------------------------------------------------
      (TRUE,  FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE,  TRUE), -- | U |
      (FALSE, TRUE,  FALSE, FALSE, FALSE, FALSE, FALSE, FALSE,  TRUE), -- | X |
      (FALSE, FALSE,  TRUE, FALSE, FALSE, FALSE,  TRUE, FALSE,  TRUE), -- | 0 |
      (FALSE, FALSE, FALSE,  TRUE, FALSE, FALSE, FALSE,  TRUE,  TRUE), -- | 1 |
      (FALSE, FALSE, FALSE, FALSE, TRUE,  FALSE, FALSE, FALSE,  TRUE), -- | Z |
      (FALSE, FALSE, FALSE, FALSE, FALSE, TRUE,  FALSE, FALSE,  TRUE), -- | W |
      (FALSE, FALSE,  TRUE, FALSE, FALSE, FALSE,  TRUE, FALSE,  TRUE), -- | L |
      (FALSE, FALSE, FALSE,  TRUE, FALSE, FALSE, FALSE,  TRUE,  TRUE), -- | H |
      ( TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE,  TRUE)  -- | - |
      );

  function MATCH (L, R: STD_ULOGIC) return BOOLEAN is
    variable VALUE: STD_ULOGIC;
  begin
    return MATCH_TABLE(L, R);
  end MATCH;

  function MATCH (L, R: STD_LOGIC_VECTOR) return BOOLEAN is
    alias LV: STD_LOGIC_VECTOR(1 to L'LENGTH) is L;
    alias RV: STD_LOGIC_VECTOR(1 to R'LENGTH) is R;
  begin
    if ((L'LENGTH < 1) or (R'LENGTH < 1)) then
      assert NO_WARNING
          report "MATCH: null detected, returning FALSE"
          severity WARNING;
      return FALSE;
    end if;
    if LV'LENGTH /= RV'LENGTH then
      assert NO_WARNING
          report "MATCH: L'LENGTH /= R'LENGTH, returning FALSE"
          severity WARNING;
      return FALSE;
    else
      for I in LV'LOW to LV'HIGH loop
        if not (MATCH_TABLE(LV(I), RV(I))) then
          return FALSE;
        end if;
      end loop;
      return TRUE;
    end if;
  end MATCH;

  -- Id: M.5
  function MATCH (L, R: STD_ULOGIC_VECTOR) return BOOLEAN is
    alias LV: STD_ULOGIC_VECTOR(1 to L'LENGTH) is L;
    alias RV: STD_ULOGIC_VECTOR(1 to R'LENGTH) is R;
  begin
    if ((L'LENGTH < 1) or (R'LENGTH < 1)) then
      assert NO_WARNING
          report "MATCH: null detected, returning FALSE"
          severity WARNING;
      return FALSE;
    end if;
    if LV'LENGTH /= RV'LENGTH then
      assert NO_WARNING
          report "MATCH: L'LENGTH /= R'LENGTH, returning FALSE"
          severity WARNING;
      return FALSE;
    else
      for I in LV'LOW to LV'HIGH loop
        if not (MATCH_TABLE(LV(I), RV(I))) then
          return FALSE;
        end if;
      end loop;
      return TRUE;
    end if;
  end MATCH;

end util;





  --============================================================================

LIBRARY ieee;
use ieee.std_logic_1164.ALL;
use ieee.std_logic_arith.ALL;
use ieee.std_logic_unsigned.ALL;
use ieee.std_logic_textio.ALL;
use std.textio.ALL;
use work.util.ALL;

entity testbench  is
	--Add entity declarations here
end testbench;


architecture behavior of  testbench is

--#COMP1 Start of component, all lines inbetween will be overwritten
--#COMP2 End of component, all lines inbetween will be overwritten

--#SIG1 Start of signal declarations, all lines inbetween will be overwritten
--#SIG2 End of signal declarations, all lines inbetween will be overwritten


begin

--#DUT1 Start of Device Under Test instantace, all lines inbetween will be overwritten
--#DUT2 End of Device Under Test instantace, all lines inbetween will be overwritten

-- *** Test Bench  **************************************************************
TB:	process
	file TestVec : Text;			--file handle
	variable FStatus : file_open_status;		--result
	variable line_in	: Line;	--Line buffer
	variable LType	: character;	--line type

	--#VAR1 Start of variables, all lines inbetween will be overwritten
	--#VAR2 End of variables, all lines inbetween will be overwritten



	procedure DoData is
	begin
	   --Assign signals	
		--#ASGN1 Start of Assignments, all lines inbetween will be overwritten
		--#ASGN2 End of Assignments, all lines inbetween will be overwritten


	   --Show test description
		if (line_in /= null) and (line_in'length /= 0) then
			write(output, "    ");		--pad with spaces so errors show better
			writeline(output, line_in);	--could do report but then get note before each line
		end if;

	   --Check outputs after delay
		--#CHK1 Start of checks, all lines inbetween will be overwritten
		--#CHK2 End of checks, all lines inbetween will be overwritten

	end DoData;


begin

	file_open(FStatus, TestVec, filename, READ_MODE);
	if FStatus /= open_ok then
		report(filename & " not found, quitting") severity error;
	else
		while (not endfile(TestVec)) loop
			readline(TestVec, line_in);
			next when line_in'length = 0;		--skip empty lines
			read(line_in, LType);				--get type of line
			case LType is
				when '*' => null;		--just a comment

				when 'D' => DoData;
				when others => report("unknown line type (1st char) of " & LType);
			end case;
		end loop;
		Deallocate(line_in);							--free old message
		line_in := new string'("      Simulation Finished");
		writeline(output, line_in);
	end if;

	file_close(TestVec);
	wait;			--All done, wait forever now
end process TB;
-- *** end Test Bench  ***

end;

--Test Vectors
--
--The test vector file controls the test process.  The first character on the test
--vector line specifies its function.  One with a * is a comment line and is
--ignored.  The only other character used is D for data.  A data line consists of:
--D <Drive> <Check> <comment>
--The Drive signal contains all the drive output states.  The check all the the
--correct results.  The end comment is displayed unless verbose is false.
--If there are decimal and/or time values these come between <Check> <comment>.
--The comment is always the last part of each line


